@@ -1,3 +1,4 @@
+import numbers
 import os
 from datetime import date
 import pandas as pd
@@ -13,6 +14,7 @@ from arcgis.geometry._types import Point
 from time import sleep
 from tqdm import tqdm
 from math import ceil
+import re
 
 try:
     import geopandas as gpd
@@ -34,6 +36,9 @@ _default_limit = 100000
 _url_error_msg = "There is likely an issue with the website. Open the URL {} with a web browser to confirm. " + \
                     "See a list of known site outages at https://github.com/openpolicedata/opd-data/blob/main/outages.csv"
 
+_last_arcgis_url = None
+_last_arcgis_date_format = None
+
 # This is for use if import data sets using Socrata. It is not required.
 # Requests made without an app_token will be subject to strict throttling limits
 # Get a App Token here: http://dev.socrata.com/docs/app-tokens.html
@@ -69,7 +74,7 @@ def load_csv(url, date_field=None, year_filter=None, agency_field=None, agency=N
         DataFrame containing table imported from CSV
     '''
     
-    if ".zip" in url or not pbar:
+    if ".zip" in url:
         with warnings.catch_warnings():
             # Perhaps use requests iter_content/iter_lines as below to read large CSVs so progress can be shown
             warnings.simplefilter("ignore", category=pd.errors.DtypeWarning)
@@ -89,7 +94,9 @@ def load_csv(url, date_field=None, year_filter=None, agency_field=None, agency=N
             raise e
         with requests.get(url, params=None, stream=True) as resp:
             try:
-                table = pd.read_csv(TqdmReader(resp), nrows=limit, encoding_errors='surrogateescape')
+                table = pd.read_csv(TqdmReader(resp, pbar=pbar), nrows=limit, encoding_errors='surrogateescape')
+            except (urllib.error.HTTPError, pd.errors.ParserError) as e:
+                raise OPD_DataUnavailableError(*e.args, _url_error_msg.format(url))
             except Exception as e:
                 raise e
 
@@ -127,8 +134,14 @@ def load_arcgis(url, date_field=None, year=None, limit=None, pbar=True):
     # For tables, it provides basic information about the table such as its ID, name, fields, types, and templates. 
     # For feature layers, in addition to the table information, it provides information such as its geometry type, min and max scales, and spatial reference.
 
-    if url[-1] == "/":
-        url = url[0:-1]
+    p = re.search(r"(MapServer|FeatureServer)/\d+", url)
+    url = url[:p.span()[1]]
+
+    global _last_arcgis_url, _last_arcgis_date_format
+    if url != _last_arcgis_url:
+        _last_arcgis_date_format = None
+        _last_arcgis_url = url
+
     last_slash = url.rindex("/")
     layer_num = url[last_slash+1:]
     base_url = url[:last_slash]
@@ -141,7 +154,17 @@ def load_arcgis(url, date_field=None, year=None, limit=None, pbar=True):
 
     # Get metadata
     r = requests.get(base_url + "/" + layer_num + "?f=pjson")
-    r.raise_for_status()
+
+    try:
+        r.raise_for_status()
+    except requests.HTTPError as e:
+        if len(e.args)>0:
+            if "503 Server Error" in e.args[0]:
+                raise OPD_DataUnavailableError(base_url, f"Layer # = {layer_num}", e.args)
+
+        else: raise e
+    except e: raise e
+    
     meta = r.json()
     if "maxRecordCount" in meta and \
         (not user_limit or (user_limit and limit > meta["maxRecordCount"])):
@@ -184,25 +207,25 @@ def load_arcgis(url, date_field=None, year=None, limit=None, pbar=True):
     
     where_query = ""
     if date_field!=None and year!=None:
-        start_date, stop_date = _process_date(year, inclusive=False)
-        
-        where_query = f"{date_field} >= '{start_date}' AND  {date_field} < '{stop_date}'"
+        where_query, record_count = _build_arcgis_where_query(base_url, layer_num, active_layer, date_field, year, _last_arcgis_date_format)
     else:
         where_query = '1=1'
-
-    try:
-        record_count = active_layer.query(where=where_query, return_count_only=True)
-        if record_count==0:
-            return None
-    except Exception as e:
-        if len(e.args)>0 and "Error Code: 429" in e.args[0]:
-            raise OPD_TooManyRequestsError(base_url, f"Layer # = {layer_num}", *e.args, _url_error_msg.format(url))
-        else:
+        try:
+            record_count = active_layer.query(where=where_query, return_count_only=True)
+        except Exception as e:
+            if len(e.args)>0 and "Error Code: 429" in e.args[0]:
+                raise OPD_TooManyRequestsError(base_url, f"Layer # = {layer_num}", *e.args, _url_error_msg.format(url))
+            else:
+                raise
+        except:
             raise
-    except:
-        raise
 
+    if record_count==0:
+        return None
+   
     if user_limit:
+        if record_count < total:
+            total = record_count
         num_batches = ceil(total / limit)
     else:
         num_batches = ceil(record_count / limit)
@@ -320,7 +343,7 @@ def load_socrata(url, data_set, date_field=None, year=None, opt_filter=None, sel
 
     where = ""
     if date_field!=None and year!=None:
-        start_date, stop_date = _process_date(year,inclusive=True, date_field=date_field)
+        start_date, stop_date = _process_date(year, date_field=date_field)
         where = date_field + " between '" + start_date + "' and '" + stop_date +"'"
 
     if opt_filter is not None:
@@ -446,7 +469,7 @@ def load_socrata(url, data_set, date_field=None, year=None, opt_filter=None, sel
     return df
 
 
-def _process_date(date, inclusive, date_field=None):
+def _process_date(date, date_field=None):
     if not isinstance(date, list):
         date = [date, date]
 
@@ -576,7 +599,7 @@ class TqdmReader:
     # Older versions of pandas check if reader has these properties even though they are not used
     write = []
     __iter__ = []
-    def __init__(self, resp, limit=None):
+    def __init__(self, resp, pbar=True, limit=None):
         total_size = int(resp.headers.get("Content-Length", 0))
 
         self.rows_read = 0
@@ -585,21 +608,24 @@ class TqdmReader:
         else:
             self.limit = float("inf")
         self.resp = resp
-        self.bar = tqdm(
-            desc=resp.url,
-            total=total_size,
-            unit="iB",
-            unit_scale=True,
-            unit_divisor=1024,
-            leave=False
-        )
+        self.pbar = pbar
+        if self.pbar:
+            self.bar = tqdm(
+                desc=resp.url,
+                total=total_size,
+                unit="iB",
+                unit_scale=True,
+                unit_divisor=1024,
+                leave=False
+            )
 
         self.reader = self.read_from_stream()
 
     def read_from_stream(self):
         for line in self.resp.iter_lines():
             line += b"\n"
-            self.bar.update(len(line))
+            if self.pbar:
+                self.bar.update(len(line))
             yield line
 
     def read(self, n=0):
@@ -610,15 +636,77 @@ class TqdmReader:
             self.rows_read += 1
             return next(self.reader)
         except StopIteration:
-            self.bar.update(self.bar.total - self.bar.n)
+            if self.pbar:
+                self.bar.update(self.bar.total - self.bar.n)
             return ""
 
+def _build_arcgis_where_query(base_url, layer_num, active_layer, date_field, year, date_format):
+    global _last_arcgis_date_format
+    record_count = 0
+    where_query = ""
+    if date_format==0 or date_format==None:
+        start_date, stop_date = _process_date(year)
+        
+        where_query = f"{date_field} >= '{start_date}' AND  {date_field} < '{stop_date}'"
+    
+        try:
+            record_count = active_layer.query(where=where_query, return_count_only=True)
+        except Exception as e:
+            if len(e.args)>0 and "Error Code: 429" in e.args[0]:
+                raise OPD_TooManyRequestsError(base_url, f"Layer # = {layer_num}", *e.args, _url_error_msg.format(url))
+            elif len(e.args)>0 and "Unable to complete operation.\n(Error Code: 400)" in e.args[0]:
+                # This query throws an error for this dataset. Try another one below
+                pass
+            else:
+                raise
+        except:
+            raise
+
+        if record_count>0 or date_format==0:
+            _last_arcgis_date_format = 0
+            return where_query, record_count
+
+    where_formats = [
+        "{} LIKE '%[0-9][0-9]/[0-9][0-9]/{}%'",   # mm/dd/yyyy
+        "{} LIKE '{}/[0-9][0-9]'",                # yyyy/mm
+        "{} = {}",                # yyyy
+    ]
+    # Make year iterable
+    year = [year] if isinstance(year, numbers.Number) else year
+
+    for format in where_formats:
+        if date_format==format or date_format==None:
+            where_query = format.format(date_field, year[0])
+            for x in year[1:]:
+                where_query = f"{where_query} or " + format.format(date_field, x)
+
+            try:
+                record_count = active_layer.query(where=where_query, return_count_only=True)
+            except Exception as e:
+                if len(e.args)>0 and "Error Code: 429" in e.args[0]:
+                    raise OPD_TooManyRequestsError(base_url, f"Layer # = {layer_num}", *e.args, _url_error_msg.format(url))
+                elif len(e.args)>0 and ("Error Code: 400" in e.args[0] or "Failed to execute query" in e.args[0]):
+                    # This query throws an error for this dataset. Try another one below
+                    pass
+                else:
+                    raise
+            except:
+                raise
+
+            if record_count>0 or date_format==format:
+                _last_arcgis_date_format = format
+                return where_query, record_count
+
+    return where_query, record_count
+
+        
 if __name__ == "__main__":
     import time
     _default_limit = 10000
     start_time = time.time()
-    url = "https://gis.charlottenc.gov/arcgis/rest/services/CMPD/CMPD/MapServer/16/"
-    url = 'https://services1.arcgis.com/zdB7qR0BtYrg0Xpl/arcgis/rest/services/ODC_CRIME_STOPS_P/FeatureServer/32'
-    date_field = 'TIME_PHONEPICKUP'
-    load_arcgis(url,date_field=date_field, year=2020)
+    url = "https://gis.charlottenc.gov/arcgis/rest/services/CMPD/CMPD/MapServer/13/"
+    # url = 'https://services1.arcgis.com/zdB7qR0BtYrg0Xpl/arcgis/rest/services/ODC_CRIME_STOPS_P/FeatureServer/32'
+    date_field = 'YR'
+    years = _get_years("ArcGIS", url, date_field)
+    load_arcgis(url, date_field, [2020,2021])
     print(f"Completed in {time.time()-start_time} seconds")
